@@ -6,6 +6,7 @@ The final `Indexer`_ classes exposes a high-level Searcher and Writer.
 
 from future_builtins import map, zip
 import os
+import re
 import itertools, operator
 import contextlib
 import abc, collections
@@ -166,6 +167,10 @@ class IndexReader(object):
     def directory(self):
         "reader's lucene Directory"
         return self.indexReader.directory()
+    @property
+    def timestamp(self):
+        "timestamp of reader's last commit"
+        return self.indexCommit.timestamp / 1000.0
     def copy(self, dest, query=None, exclude=None, optimize=False):
         """Copy the index to the destination directory.
         Optimized to use hard links if the destination is a file system path.
@@ -175,12 +180,19 @@ class IndexReader(object):
         :param exclude: optional lucene Query to exclude documents
         :param optimize: optionally optimize destination index
         """
+        filter = lucene.IndexFileNameFilter.getFilter()
+        filenames = (filename for filename in self.directory.listAll() if filter.accept(None, filename))
         if isinstance(dest, lucene.Directory):
-            lucene.Directory.copy(self.directory, dest, False)
+            src = self.directory
+            try:
+                lucene.Directory.copy(src, dest, False)
+            except lucene.InvalidArgsError:
+                for filename in filenames:
+                    src.copy(dest, filename, filename)
         else:
             src = lucene.FSDirectory.cast_(self.directory).file.path
             os.path.isdir(dest) or os.makedirs(dest)
-            for filename in set(self.directory.listAll()).difference(['write.lock']):
+            for filename in filenames:
                 os.link(os.path.join(src, filename), os.path.join(dest, filename))
         with contextlib.closing(IndexWriter(dest)) as writer:
             if query:
@@ -190,7 +202,7 @@ class IndexReader(object):
             writer.commit()
             writer.expungeDeletes()
             if optimize:
-                writer.optimize()
+                writer.optimize(optimize)
         return len(writer)
     def count(self, name, value):
         "Return number of documents with given term."
@@ -250,17 +262,13 @@ class IndexReader(object):
                 yield (doc, termdocs.freq()) if counts else doc
     def positions(self, name, value, payloads=False):
         "Generate doc ids and positions which contain given term, optionally only with payloads."
+        array = lucene.JArray_byte('')
         with contextlib.closing(self.termPositions(lucene.Term(name, value))) as termpositions:
             while termpositions.next():
                 doc = termpositions.doc()
                 positions = (termpositions.nextPosition() for n in xrange(termpositions.freq()))
                 if payloads:
-                    items, array = [], lucene.JArray_byte('')
-                    for position in positions:
-                        if termpositions.payloadAvailable:
-                            data = termpositions.getPayload(array, 0)
-                            items.append((position, data.string_ if hasattr(data, 'string_') else ''.join(data)))
-                    yield doc, items
+                    yield doc, [(position, termpositions.getPayload(array, 0).string_) for position in positions if termpositions.payloadAvailable]
                 else:
                     yield doc, list(positions)
     def comparator(self, name, type='string', parser=None):
@@ -280,17 +288,11 @@ class IndexReader(object):
         :param positions: optionally include slice positions instead of counts
         :param payloads: optionally only include slice positions with payloads
         """
-        spans = query.getSpans(self.indexReader)
-        spans = itertools.takewhile(lucene.Spans.next, itertools.repeat(spans))
+        spans = itertools.takewhile(lucene.Spans.next, itertools.repeat(query.getSpans(self.indexReader)))
         for doc, spans in itertools.groupby(spans, key=lucene.Spans.doc):
             if payloads:
-                items = []
-                for span in spans:
-                    if span.payloadAvailable:
-                        payloads = map(lucene.JArray_byte.cast_, span.payload)
-                        payloads = [data.string_ if hasattr(data, 'string_') else ''.join(data) for data in payloads]
-                        items.append((span.start(), span.end(), payloads))
-                yield doc, items
+                yield doc, [(span.start(), span.end(), [lucene.JArray_byte.cast_(data).string_ for data in span.payload]) \
+                    for span in spans if span.payloadAvailable]
             elif positions:
                 yield doc, [(span.start(), span.end()) for span in spans]
             else:
@@ -319,6 +321,17 @@ class IndexReader(object):
         for name, value in attrs.items():
             setattr(mlt, name, value)
         return mlt.like(lucene.StringReader(doc) if isinstance(doc, basestring) else doc)
+    def overlap(self, left, right):
+        "Return intersection count of cached filters."
+        count = 0
+        for reader in self.sequentialSubReaders:
+            docsets = left.getDocIdSet(reader), right.getDocIdSet(reader)
+            try:
+                count += lucene.OpenBitSet.intersectionCount(*docsets)
+            except lucene.InvalidArgsError: # verify docsets are either cached or empty
+                if not all(lucene.OpenBitSet.instance_(docset) or docset.iterator().nextDoc() == lucene.Integer.MAX_VALUE for docset in docsets):
+                    raise
+        return int(count)
 
 class Searcher(object):
     "Mixin interface common among searchers."
@@ -353,7 +366,7 @@ class Searcher(object):
         other.shared, other.owned = self.shared, closing([reader])
         other.filters.update((key, value if isinstance(value, lucene.Filter) else dict(value)) for key, value in self.filters.items())
         if filters:
-            other.facets([], *other.filters)
+            other.facets(Query.any(), *other.filters)
         other.sorters = dict(self.sorters)
         if sorters:
             for field in self.sorters:
@@ -450,19 +463,25 @@ class Searcher(object):
             stats = topdocs.totalHits, topdocs.maxScore
         stats *= not isinstance(timeout, lucene.JavaError)
         return Hits(self, ids, scores, *stats)
-    def facets(self, ids, *keys):
+    def facets(self, query, *keys):
         """Return mapping of document counts for the intersection with each facet.
         
-        :param ids: document ids or lucene Filter
+        :param ids: query string, lucene Query, or lucene Filter
         :param keys: field names, term tuples, or any keys to previously cached filters
         """
         counts = collections.defaultdict(dict)
-        if not isinstance(ids, lucene.Filter):
-            ids = Filter(ids)
+        if isinstance(query, basestring):
+            query = self.parse(query)
+        if isinstance(query, lucene.Query):
+            query = lucene.QueryWrapperFilter(query)
+        if not isinstance(query, lucene.Filter):
+            query = Filter(query)
+        elif not isinstance(query, lucene.CachingWrapperFilter):
+            query = lucene.CachingWrapperFilter(query)
         for key in keys:
             filters = self.filters.get(key)
             if isinstance(filters, lucene.Filter):
-                counts[key] = Filter.overlap(ids, filters, self.indexReader)
+                counts[key] = self.overlap(query, filters)
             else:
                 name, value = (key, None) if isinstance(key, basestring) else key
                 filters = self.filters.setdefault(name, {})
@@ -473,7 +492,7 @@ class Searcher(object):
                 for value in values:
                     if value not in filters:
                         filters[value] = Query.term(name, value).filter()
-                    counts[name][value] = Filter.overlap(ids, filters[value], self.indexReader)
+                    counts[name][value] = self.overlap(query, filters[value])
         return dict(counts)
     def sorter(self, field, type='string', parser=None, reverse=False):
         "Return `SortField`_ with cached attributes if available."
@@ -508,6 +527,17 @@ class Searcher(object):
             words = set(self.terms(field, text, minSimilarity=minSimilarity)) - corrections
             for word in sorted(words, key=spellchecker.__getitem__, reverse=True):
                 yield word
+    def match(self, document, *queries):
+        "Generate scores for all queries against a given document mapping."
+        searcher = lucene.MemoryIndex()
+        for name, value in document.items():
+            if isinstance(value, basestring):
+                value = value, self.analyzer
+            elif isinstance(value, lucene.TokenStream):
+                value = value,
+            searcher.addField(name, *value)
+        for query in queries:
+            yield searcher.search(self.parse(query))
 
 class IndexSearcher(Searcher, lucene.IndexSearcher, IndexReader):
     """Inherited lucene IndexSearcher, with a mixed-in IndexReader.
@@ -543,6 +573,14 @@ class MultiSearcher(Searcher, lucene.MultiSearcher, IndexReader):
             self.indexReader = lucene.MultiReader(list(map(operator.attrgetter('indexReader'), searchers)), False)
             self.owned = closing([self.indexReader])
         Searcher.__init__(self, searchers, analyzer)
+    @property
+    def version(self):
+        return ' '.join(str(reader.version) for reader in self.sequentialSubReaders)
+    @property
+    def timestamp(self):
+        return max(IndexReader(reader).timestamp for reader in self.sequentialSubReaders)
+    def overlap(self, *filters):
+        return sum(IndexReader(reader).overlap(*filters) for reader in self.sequentialSubReaders)
 
 class ParallelMultiSearcher(MultiSearcher, lucene.ParallelMultiSearcher):
     "Inherited lucene ParallelMultiSearcher."
@@ -566,17 +604,14 @@ class IndexWriter(lucene.IndexWriter):
         if not isinstance(directory, lucene.Directory):
             directory = lucene.RAMDirectory() if directory is None else lucene.FSDirectory.open(lucene.File(directory))
             self.shared.add(directory)
-        mfl = lucene.IndexWriter.MaxFieldLength.LIMITED
-        if mode == 'a':
-            lucene.IndexWriter.__init__(self, directory, analyzer, mfl)
-        else:
-            lucene.IndexWriter.__init__(self, directory, analyzer, bool('rw'.index(mode)), mfl)
+        args = [] if mode == 'a' else [bool('rw'.index(mode))]
+        args.append(lucene.IndexWriter.MaxFieldLength.LIMITED)
+        lucene.IndexWriter.__init__(self, directory, analyzer, *args)
         self.fields = {}
     @property
     def segments(self):
         "segment filenames with document counts"
-        items = (seg.lower().split(':c') for seg in self.segString().split())
-        return dict((name, int(value)) for name, value in items)
+        return dict((name, int(value)) for name, value in re.findall('(\w+).*:.x?(\d+)', self.segString()))
     def set(self, name, cls=Field, **params):
         """Assign parameters to field name.
         
@@ -585,41 +620,40 @@ class IndexWriter(lucene.IndexWriter):
         :param params: store,index,termvector options compatible with `Field`_
         """
         self.fields[name] = cls(name, **params)
-    def add(self, document=(), boost=1.0, **terms):
-        """Add document to index.
-        Document is comprised of name: value pairs, where the values may be one or multiple strings.
-        
-        :param document: optional document terms as a dict or items
-        :param terms: additional terms to document
-        """
-        terms.update(document)
+    def document(self, document=(), **terms):
+        "Return lucene Document from mapping of field names to one or multiple values."
         doc = lucene.Document()
-        doc.boost = boost
-        for name, values in terms.items():
+        for name, values in dict(document, **terms).items():
             if isinstance(values, Atomic):
                 values = values,
             for field in self.fields[name].items(*values):
                 doc.add(field)
+        return doc
+    def add(self, document=(), boost=1.0, **terms):
+        "Add :meth:`document` to index with optional boost."
+        doc = self.document(document, **terms)
+        doc.boost = boost
         self.addDocument(doc)
+    def update(self, name, value='', document=(), boost=1.0, **terms):
+        "Atomically delete documents which match given term and add the new :meth:`document` with optional boost."
+        doc = self.document(document, **terms)
+        doc.boost = boost
+        self.updateDocument(lucene.Term(name, *([value] if value else doc.getValues(name))), doc)
     def delete(self, *query, **options):
         """Remove documents which match given query or term.
         
         :param query: :meth:`Searcher.search` compatible query, or optimally a name and value
         :param options: additional :meth:`Analyzer.parse` options
         """
-        if len(query) == 1:
-            self.deleteDocuments(self.parse(*query, **options))
-        else:
-            self.deleteDocuments(lucene.Term(*query))
+        parse = self.parse if len(query) == 1 else lucene.Term
+        self.deleteDocuments(parse(*query, **options))
     def __iadd__(self, directory):
         "Add directory (or reader, searcher, writer) to index."
         if isinstance(directory, basestring):
             with contextlib.closing(lucene.FSDirectory.open(lucene.File(directory))) as directory:
                 self.addIndexesNoOptimize([directory])
         else:
-            if not isinstance(directory, lucene.Directory):
-                directory = directory.directory
-            self.addIndexesNoOptimize([directory])
+            self.addIndexesNoOptimize([getattr(directory, 'directory', directory)])
         return self
 
 class Indexer(IndexWriter):
@@ -628,6 +662,7 @@ class Indexer(IndexWriter):
     """
     def __init__(self, *args, **kwargs):
         IndexWriter.__init__(self, *args, **kwargs)
+        IndexWriter.commit(self)
         self.indexSearcher = IndexSearcher(self.directory, self.analyzer)
     def __getattr__(self, name):
         if name == 'indexSearcher':
@@ -639,10 +674,18 @@ class Indexer(IndexWriter):
         return iter(self.indexSearcher)
     def __getitem__(self, id):
         return self.indexSearcher[id]
-    def commit(self, **caches):
+    def commit(self, expunge=False, optimize=False, **caches):
         """Commit writes and refresh searcher.  Not thread-safe.
         
+        :param expunge: expunge deletes
+        :param optimize: optimize index, optionally supply number of segments
         :param caches: :meth:`Searcher.reopen` caches to refresh
         """
         IndexWriter.commit(self)
+        if expunge:
+            self.expungeDeletes()
+            IndexWriter.commit(self)
+        if optimize:
+            self.optimize(optimize)
+            IndexWriter.commit(self)
         self.indexSearcher = self.indexSearcher.reopen(**caches)
