@@ -9,6 +9,7 @@ CherryPy and Lucene VM integration issues:
 """
 
 from future_builtins import filter, map
+import warnings
 import re
 import time
 import httplib
@@ -16,6 +17,7 @@ import threading
 import collections
 import itertools, operator
 import os, optparse
+from email.utils import formatdate
 from contextlib import contextmanager
 try:
     import simplejson as json
@@ -33,23 +35,36 @@ def tool(hook):
     return decorator
 
 @tool('before_handler')
-def json_(indent=None):
-    """Transform responses into json format.
-    Specify tools.json.indent for pretty printing.
+def json_(indent=None, content_type='application/json', process_body=None):
+    """Handle request bodies and responses in json format.
+
+    :param indent: indentation level for pretty printing
+    :param content_type: response content-type header
+    :param process_body: optional function to process body into request.params
     """
-    handler = cherrypy.request.handler
+    request = cherrypy.serving.request
+    headers = cherrypy.response.headers
+    if request.headers.get('content-type', '').endswith('json'):
+        with HTTPError(httplib.BAD_REQUEST, ValueError):
+            request.json = json.load(request.body)
+        if process_body is not None:
+            with HTTPError(httplib.BAD_REQUEST, TypeError):
+                request.params.update(process_body(request.json))
+    elif request.headers.get('content-type') == 'application/x-www-form-urlencoded':
+        headers['Warning'] = '199 lupyne "Content-Type application/x-www-form-urlencoded has been deprecated and replaced with application/json"'
+    handler = request.handler
     def json_handler(*args, **kwargs):
         body = handler(*args, **kwargs)
-        if cherrypy.response.headers['content-type'].startswith('text/'):
-            cherrypy.response.headers['content-type'] = 'text/x-json'
+        if headers['content-type'].startswith('text/'):
+            headers['content-type'] = content_type
             body = json.dumps(body, indent=indent)
         return body
-    cherrypy.request.handler = json_handler
+    request.handler = json_handler
 
 @tool('on_start_resource')
 def allow(methods=('GET', 'HEAD')):
-    "Only allow methods specified in tools.allow.methods."
-    request = cherrypy.request
+    "Only allow specified methods."
+    request = cherrypy.serving.request
     if request.method not in methods and not isinstance(request.handler, cherrypy.HTTPError):
         cherrypy.response.headers['allow'] = ', '.join(methods)
         message = "The path {0!r} does not allow {1}.".format(request.path_info, request.method)
@@ -58,31 +73,57 @@ def allow(methods=('GET', 'HEAD')):
 @tool('before_finalize')
 def time_():
     "Return response time in headers."
-    cherrypy.response.headers['x-response-time'] = time.time() - cherrypy.response.time
+    response = cherrypy.serving.response
+    response.headers['x-response-time'] = time.time() - response.time
+
+@tool('on_start_resource')
+def validate(methods=('GET', 'HEAD'), etag=True, last_modified=True, max_age=None, expires=None):
+    """Return and validate caching headers for GET requests.
+
+    :param methods: only set headers for specified methods
+    :param etag: return weak entity tag based on index version and validate if-match headers
+    :param last_modified: return last-modified based on index timestamp and validate if-modified headers
+    :param max_age: return cache-control max-age and age header based on last update
+    :param expires: return expires header based on last update
+    """
+    request = cherrypy.serving.request
+    headers = cherrypy.response.headers
+    if request.method in methods and not isinstance(request.handler, cherrypy.HTTPError):
+        if etag:
+            headers['etag'] = 'W/"{0}"'.format(request.app.root.searcher.version)
+            cherrypy.lib.cptools.validate_etags()
+        if last_modified:
+            headers['last-modified'] = formatdate(request.app.root.searcher.timestamp, usegmt=True)
+            cherrypy.lib.cptools.validate_since()
+        if max_age is not None:
+            headers['age'] = int(time.time() - request.app.root.updated)
+            headers['cache-control'] = 'max-age={0}'.format(max_age)
+        if expires is not None:
+            headers['expires'] = formatdate(expires + request.app.root.updated, usegmt=True)
 
 def json_error(version, **body):
     "Transform errors into json format."
-    cherrypy.response.headers['content-type'] = 'text/x-json'
-    return json.dumps(body, indent=2)
+    tool = cherrypy.request.toolmaps['tools'].get('json', {})
+    cherrypy.response.headers['content-type'] = tool.get('content_type', 'application/json')
+    return json.dumps(body, indent=tool.get('indent'))
 
-def attach_thread(id):
+def attach_thread(id=None):
     "Attach current cherrypy worker thread to lucene VM."
     lucene.getVMEnv().attachCurrentThread()
 
 class Autoreloader(cherrypy.process.plugins.Autoreloader):
     "Autoreload monitor compatible with lucene VM."
     def run(self):
-        lucene.getVMEnv().attachCurrentThread()
+        attach_thread()
         cherrypy.process.plugins.Autoreloader.run(self)
 
 class AttachedMonitor(cherrypy.process.plugins.Monitor):
     "Periodically run a callback function in an attached thread."
     def __init__(self, bus, callback, frequency=cherrypy.process.plugins.Monitor.frequency):
-        cherrypy.process.plugins.Monitor.__init__(self, bus, self.run, frequency)
-        self._callback = callback
-    def run(self):
-        lucene.getVMEnv().attachCurrentThread()
-        self._callback()
+        def run():
+            attach_thread()
+            callback()
+        cherrypy.process.plugins.Monitor.__init__(self, bus, run, frequency)
 
 @contextmanager
 def HTTPError(status, *exceptions):
@@ -94,18 +135,30 @@ def HTTPError(status, *exceptions):
 
 class WebSearcher(object):
     "Dispatch root with a delegated Searcher."
-    _cp_config = {'tools.json.on': True, 'tools.allow.on': True, 'tools.time.on': True, 'error_page.default': json_error,
-        'tools.gzip.on': True, 'tools.gzip.mime_types': ['text/html', 'text/plain', 'text/x-json']}
+    _cp_config = dict.fromkeys(map('tools.{0}.on'.format, ['gzip', 'json', 'allow', 'time', 'validate']), True)
+    _cp_config.update({'error_page.default': json_error, 'tools.gzip.mime_types': ['text/html', 'text/plain', 'application/json']})
     def __init__(self, *directories, **kwargs):
-        self.indexer = engine.MultiSearcher(directories, **kwargs) if len(directories) > 1 else engine.IndexSearcher(*directories, **kwargs)
+        self.searcher = engine.MultiSearcher(directories, **kwargs) if len(directories) > 1 else engine.IndexSearcher(*directories, **kwargs)
+        self.updated = time.time()
+    @classmethod
+    def new(cls, *args, **kwargs):
+        "Return new uninitialized root which can be mounted on dispatch tree before VM initialization."
+        self = object.__new__(cls)
+        self.args, self.kwargs = args, kwargs
+        return self
+    def init(self, vmargs='-Xrs', **kwargs):
+        "Callback to initialize VM and root object after daemonizing."
+        lucene.initVM(vmargs=vmargs, **kwargs)
+        self.__init__(*self.__dict__.pop('args'), **self.__dict__.pop('kwargs'))
     def close(self):
-        self.indexer.close()
-    def parse(self, q, **options):
+        self.searcher.close()
+    @staticmethod
+    def parse(searcher, q, **options):
         "Return parsed query using q.* parser options."
         options = dict((key.partition('.')[-1], options[key]) for key in options if key.startswith('q.'))
         field = options.pop('field', [])
         fields = [field] if isinstance(field, basestring) else field
-        fields = [re.match('(\w+)\^?([\d\.]*)', name).groups() for name in fields]
+        fields = [name.partition('^')[::2] for name in fields]
         if any(boost for name, boost in fields):
             field = dict((name, float(boost or 1.0)) for name, boost in fields)
         elif isinstance(field, basestring):
@@ -118,8 +171,11 @@ class WebSearcher(object):
         for key in set(options) - set(['op', 'version']):
             with HTTPError(httplib.BAD_REQUEST, ValueError):
                 options[key] = json.loads(options[key])
-        return q and self.indexer.parse(q, field=field, **options)
-    def select(self, fields=None, **options):
+        if q is not None:
+            with HTTPError(httplib.BAD_REQUEST, lucene.JavaError):
+                return searcher.parse(q, field=field, **options)
+    @staticmethod
+    def select(fields=None, **options):
         "Return parsed field selectors: stored, multi-valued, and indexed."
         if fields is not None:
             fields = dict.fromkeys(filter(None, fields.split(',')))
@@ -132,17 +188,23 @@ class WebSearcher(object):
     @cherrypy.expose
     @cherrypy.tools.allow(methods=['POST'])
     def refresh(self, **caches):
-        """Refresh index.
+        raise cherrypy.HTTPRedirect(cherrypy.request.script_name + '/update', httplib.MOVED_PERMANENTLY)
+    @cherrypy.expose
+    @cherrypy.tools.json(process_body=dict.fromkeys)
+    @cherrypy.tools.allow(methods=['POST'])
+    def update(self, **caches):
+        """Refresh index version.
         
-        **POST** /refresh
+        **POST** /update
             Reopen searcher, optionally reloading caches, and return document count.
             
-            &filters&sorters&spellcheckers
+            ["filters"|"sorters"|"spellcheckers",... ]
             
             :return: *int*
         """
-        self.indexer = self.indexer.reopen(**dict.fromkeys(caches, True))
-        return len(self.indexer)
+        self.searcher = self.searcher.reopen(**dict.fromkeys(caches, True))
+        self.updated = time.time()
+        return len(self.searcher)
     @cherrypy.expose
     def index(self):
         """Return index information.
@@ -150,13 +212,14 @@ class WebSearcher(object):
         **GET** /
             Return a mapping of the directory to the document count.
             
-            :return: {*string*: *int*}
+            :return: {*string*: *int*,... }
         """
-        if isinstance(self.indexer, lucene.MultiSearcher):
-            return dict((unicode(reader.directory()), reader.numDocs()) for reader in self.indexer.sequentialSubReaders)
-        return {unicode(self.indexer.directory): len(self.indexer)}
+        searcher = self.searcher
+        if isinstance(searcher, lucene.MultiSearcher):
+            return dict((unicode(reader.directory()), reader.numDocs()) for reader in searcher.sequentialSubReaders)
+        return {unicode(searcher.directory): len(searcher)}
     @cherrypy.expose
-    def docs(self, id=None, fields=None, **options):
+    def docs(self, *path, **options):
         """Return ids or documents.
         
         **GET** /docs
@@ -164,21 +227,24 @@ class WebSearcher(object):
             
             :return: [*int*,... ]
         
-        **GET** /docs/*int*?
-            Return document mappings, optionally selecting stored, multi-valued, and cached indexed fields.
+        **GET** /docs/[*int*\|\ *chars*/*chars*]?
+            Return document mapping from id or unique name and value.
+            Optionally select stored, multi-valued, and cached indexed fields.
             
             &fields=\ *chars*,... &fields.multi=\ *chars*,... &fields.indexed=\ *chars*\ [:*chars*],...
             
             :return: {*string*: *string*\|\ *array*,... }
         """
-        if id is None:
-            return list(self.indexer)
-        fields, multi, indexed = self.select(fields, **options)
-        with HTTPError(httplib.NOT_FOUND, ValueError, lucene.JavaError):
-            id = int(id)
-            doc = self.indexer[id] if fields is None else self.indexer.get(id, *itertools.chain(fields, multi))
+        searcher = self.searcher
+        if not path:
+            return list(searcher)
+        fields, multi, indexed = self.select(**options)
+        with HTTPError(httplib.NOT_FOUND, ValueError):
+            id, = map(int, path) if len(path) == 1 else searcher.docs(*path)
+        with HTTPError(httplib.NOT_FOUND, lucene.JavaError):
+            doc = searcher[id] if fields is None else searcher.get(id, *itertools.chain(fields, multi))
         result = doc.dict(*multi, **(fields or {}))
-        result.update((item[0], self.indexer.comparator(*item)[id]) for item in indexed)
+        result.update((item[0], searcher.comparator(*item)[id]) for item in indexed)
         return result
     @cherrypy.expose
     def search(self, q=None, count=None, start=0, fields=None, sort=None, facets='', group='', hl='', mlt=None, spellcheck=0, timeout=None, **options):
@@ -198,7 +264,7 @@ class WebSearcher(object):
                 maximum number of docs to return and offset to start at
             
             &fields=\ *chars*,... &fields.multi=\ *chars*,... &fields.indexed=\ *chars*\ [:*chars*],...
-                only include selected stored fields; multi-valued fields returned in an array; indexed fields are cached
+                only include selected stored fields; multi-valued fields returned in an array; indexed fields with optional type are cached
             
             &sort=\ [-]\ *chars*\ [:*chars*],... &sort.scores[=max]
                 | field name, optional type, minus sign indicates descending
@@ -207,8 +273,9 @@ class WebSearcher(object):
             &facets=\ *chars*,...
                 include facet counts for given field names; facets filters are cached
             
-            &group=\ *chars*\ &group.count=1
-                group documents by field value, up to given maximum count
+            &group=\ *chars*\ [:*chars*]&group.count=1&group.limit=\ *int*
+                | group documents by field value with optional type, up to given maximum count
+                | limit number of groups which return docs
             
             &hl=\ *chars*,... &hl.count=1&hl.tag=strong&hl.enable=[fields|terms]
                 | stored fields to return highlighted
@@ -224,20 +291,21 @@ class WebSearcher(object):
                 | maximum number of spelling corrections to return for each query term, grouped by field
                 | original query is still run; use q.spellcheck=true to affect query parsing
             
-            &timeout=\ *float*
+            &timeout=\ *number*
                 timeout search after elapsed number of seconds
             
             :return:
                 | {
                 | "query": *string*,
                 | "count": *int*\|null,
-                | "maxscore": *float*\|null,
-                | "docs": [{"__id__": *int*, "__score__": *float*, "__highlights__": {*string*: *array*,... }, *string*: *string*\|\ *array*,... },... ],
+                | "maxscore": *number*\|null,
+                | "docs": [{"__id__": *int*, "__score__": *number*, "__highlights__": {*string*: *array*,... }, *string*: *string*\|\ *array*,... },... ],
                 | "facets": {*string*: {*string*: *int*,... },... },
                 | "groups": [{"count": *int*, "value": *value*, "docs": [{... },... ]},... ]
                 | "spellcheck": {*string*: {*string*: [*string*,... ],... },... },
                 | }
         """
+        searcher = self.searcher
         with HTTPError(httplib.BAD_REQUEST, ValueError):
             start = int(start)
             if count is not None:
@@ -246,11 +314,11 @@ class WebSearcher(object):
             if timeout is not None:
                 timeout = float(timeout)
             gcount = int(options.get('group.count', 1))
+            glimit = int(options['group.limit']) if 'group.limit' in options else float('inf')
             hlcount = int(options.get('hl.count', 1))
             if mlt is not None:
                 mlt = int(mlt)
         reverse = False
-        searcher = getattr(self.indexer, 'indexSearcher', self.indexer)
         if sort is not None:
             sort = (re.match('(-?)(\w+):?(\w*)', field).groups() for field in sort.split(','))
             sort = [(name, (type or 'string'), (reverse == '-')) for reverse, name, type in sort]
@@ -262,10 +330,10 @@ class WebSearcher(object):
             else:
                 with HTTPError(httplib.BAD_REQUEST, AttributeError):
                     sort = [searcher.sorter(name, type, reverse=reverse) for name, type, reverse in sort]
-        q = self.parse(q, **options)
+        q = self.parse(searcher, q, **options)
         qfilter = options.pop('filter', None)
         if qfilter is not None and qfilter not in searcher.filters:
-            searcher.filters[qfilter] = engine.Query.__dict__['filter'](self.parse(qfilter, **options))
+            searcher.filters[qfilter] = engine.Query.__dict__['filter'](self.parse(searcher, qfilter, **options))
         qfilter = searcher.filters.get(qfilter)
         if mlt is not None:
             if q is not None:
@@ -295,12 +363,13 @@ class WebSearcher(object):
         docs = []
         groups = collections.defaultdict(lambda: {'docs': [], 'count': 0, 'index': len(groups)})
         if group:
-            group = searcher.comparator(*group.split(':'))
+            with HTTPError(httplib.BAD_REQUEST, AttributeError):
+                group = searcher.comparator(*group.split(':'))
             ids, scores = [], []
             for id, score in hits.items():
                 item = groups[group[id]]
                 item['count'] += 1
-                if item['count'] <= gcount:
+                if item['count'] <= gcount and item['index'] < glimit:
                     ids.append(id)
                     scores.append(score)
             hits.ids, hits.scores = ids, scores
@@ -319,14 +388,14 @@ class WebSearcher(object):
         q = q or lucene.MatchAllDocsQuery()
         if facets:
             facets = (tuple(facet.split(':')) if ':' in facet else facet for facet in facets.split(','))
-            result['facets'] = searcher.facets(engine.Query.__dict__['filter'](q), *facets)
+            result['facets'] = searcher.facets(q, *facets)
         if spellcheck:
             terms = result['spellcheck'] = collections.defaultdict(dict)
             for name, value in engine.Query.__dict__['terms'](q):
                 terms[name][value] = list(itertools.islice(searcher.correct(name, value), spellcheck))
         return result
     @cherrypy.expose
-    def terms(self, name='', value=':', docs='', counts='', **options):
+    def terms(self, name='', value=':', *path, **options):
         """Return data about indexed terms.
         
         **GET** /terms?
@@ -336,12 +405,12 @@ class WebSearcher(object):
             
             :return: [*string*,... ]
         
-        **GET** /terms/*chars*
-            Return term values for given field name.
+        **GET** /terms/*chars*\[:int|float\]?step=0
+            Return term values for given field name, with optional type and step for numeric encoded values.
             
             :return: [*string*,... ]
         
-        **GET** /terms/*chars*/*chars*\[\*\|?\|:*chars*\|~\ *float*\]
+        **GET** /terms/*chars*/*chars*\[\*\|?\|:*chars*\|~\ *number*\]
             Return term values (wildcards, slices, or fuzzy terms) for given field name.
             
             :return: [*string*,... ]
@@ -372,70 +441,128 @@ class WebSearcher(object):
             
             :return: [[*int*, [*int*,... ]],... ]
         """
+        searcher = self.searcher
         if not name:
-            return sorted(self.indexer.names(**options))
+            return sorted(searcher.names(**options))
+        if ':' in name:
+            with HTTPError(httplib.BAD_REQUEST, ValueError, AttributeError):
+                name, type = name.split(':')
+                type = getattr(__builtins__, type)
+                step = int(options.get('step', 0))
+            return list(searcher.numbers(name, step=step, type=type))
         if ':' in value:
-            start, stop = value.split(':')
-            return list(self.indexer.terms(name, start, stop or None))
+            with HTTPError(httplib.BAD_REQUEST, ValueError):
+                start, stop = value.split(':')
+            return list(searcher.terms(name, start, stop or None))
         if 'count' in options:
             with HTTPError(httplib.BAD_REQUEST, ValueError):
                 count = int(options['count'])
             if value.endswith('*'):
-                return self.indexer.suggest(name, value.rstrip('*'), count)
+                return searcher.suggest(name, value.rstrip('*'), count)
             if value.endswith('~'):
-                return list(itertools.islice(self.indexer.correct(name, value.rstrip('~')), count))
+                return list(itertools.islice(searcher.correct(name, value.rstrip('~')), count))
         if '*' in value or '?' in value:
-            return list(self.indexer.terms(name, value))
+            return list(searcher.terms(name, value))
         if '~' in value:
-            value, similarity = value.split('~')
             with HTTPError(httplib.BAD_REQUEST, ValueError):
+                value, similarity = value.split('~')
                 similarity = float(similarity or 0.5)
-            return list(self.indexer.terms(name, value, minSimilarity=similarity))
-        if not docs:
-            return self.indexer.count(name, value)
-        if docs == 'docs':
-            if counts == 'positions':
-                return list(self.indexer.positions(name, value))
-            if counts in ('', 'counts'):
-                return list(self.indexer.docs(name, value, counts=bool(counts)))
+            return list(searcher.terms(name, value, minSimilarity=similarity))
+        if not path:
+            return searcher.count(name, value)
+        if path[0] == 'docs':
+            if path[1:] == ():
+                return list(searcher.docs(name, value))
+            if path[1:] == ('counts',):
+                return list(searcher.docs(name, value, counts=True))
+            if path[1:] == ('positions',):
+                return list(searcher.positions(name, value))
         raise cherrypy.NotFound()
 
 class WebIndexer(WebSearcher):
     "Dispatch root which extends searcher to include write methods."
+    commit = WebSearcher.refresh
     def __init__(self, *args, **kwargs):
         self.indexer = engine.Indexer(*args, **kwargs)
+        self.updated = time.time()
         self.lock = threading.Lock()
+    @property
+    def searcher(self):
+        return self.indexer.indexSearcher
+    def close(self):
+        self.indexer.close()
+        WebSearcher.close(self)
     @cherrypy.expose
-    @cherrypy.tools.allow(methods=['POST'])
-    def refresh(self, **caches):
-        raise cherrypy.HTTPRedirect(cherrypy.request.script_name + '/commit', httplib.MOVED_PERMANENTLY)
-    @cherrypy.expose
-    @cherrypy.tools.allow(methods=['POST'])
-    def commit(self, **caches):
-        """Commit index changes.
+    @cherrypy.tools.json(process_body=lambda body: {'directories': list(body)})
+    @cherrypy.tools.allow(methods=['GET', 'HEAD', 'POST'])
+    def index(self, directories=()):
+        """Add indexes.  See :meth:`WebSearcher.index` for GET method.
         
-        **POST** /commit
-            Commit write operations and return document count.  See :meth:`WebSearcher.refresh` for caching options.
+        **POST** /
+            Add indexes without optimization.
+            
+            [*string*,... ]
+        """
+        if cherrypy.request.method == 'POST':
+            for directory in directories:
+                self.indexer += directory
+            cherrypy.response.status = httplib.ACCEPTED
+        return {unicode(self.indexer.directory): len(self.indexer)}
+    @cherrypy.expose
+    @cherrypy.tools.json(process_body=dict.fromkeys)
+    @cherrypy.tools.allow(methods=['POST'])
+    def update(self, **options):
+        """Commit index changes and refresh index version.
+        
+        **POST** /update
+            Commit write operations and return document count.  See :meth:`WebSearcher.update` for caching options.
+            
+            ["expunge"|"optimize",... ]
+            
+            :return: *int*
         """
         with self.lock:
-            self.indexer.commit(**dict.fromkeys(caches, True))
+            self.indexer.commit(**dict.fromkeys(options, True))
+        self.updated = time.time()
         return len(self.indexer)
     @cherrypy.expose
-    @cherrypy.tools.allow(methods=['GET', 'HEAD', 'POST'])
-    def docs(self, id=None, docs='[]', fields=None, **options):
+    @cherrypy.tools.allow(methods=['GET', 'HEAD', 'POST', 'PUT', 'DELETE'])
+    def docs(self, *path, **options):
         """Add or return documents.  See :meth:`WebSearcher.docs` for GET method.
         
         **POST** /docs
             Add documents to index.
             
-            docs=[{*string*: *string*\|\ *array*,... },... ]
+            [{*string*: *string*\|\ *array*,... },... ]
+        
+        **PUT, DELETE** /docs/*chars*/*chars*
+            Set or delete document.  Unique term should be indexed and is added to the new document.
+            
+            {*string*: *string*\|\ *array*,... }
         """
-        if cherrypy.request.method != 'POST':
-            return WebSearcher.docs(self, id, fields, **options)
-        with HTTPError(httplib.BAD_REQUEST, ValueError):
-            docs = json.loads(docs)
-        for doc in docs:
-            self.indexer.add(doc)
+        with HTTPError(httplib.NOT_FOUND, IndexError):
+            allow([('GET', 'HEAD', 'POST'), ('GET', 'HEAD'), ('GET', 'HEAD', 'PUT', 'DELETE')][len(path)])
+        request = cherrypy.serving.request
+        if request.method in ('GET', 'HEAD'):
+            return WebSearcher.docs(self, *path, **options)
+        if request.method == 'DELETE':
+            self.indexer.delete(*path)
+        elif request.method == 'PUT':
+            name, value = path
+            doc = getattr(request, 'json', {})
+            with HTTPError(httplib.CONFLICT, KeyError, AssertionError):
+                assert self.indexer.fields[name].index != lucene.Field.Index.NO, 'unique field must be indexed'
+                assert doc.setdefault(name, value) == value, 'multiple values for unique field'
+            self.indexer.update(name, value, doc)
+        else:
+            docs = getattr(request, 'json', ())
+            with HTTPError(httplib.BAD_REQUEST, KeyError, ValueError): # deprecated
+                if isinstance(docs, dict):
+                    docs = docs['docs']
+                if 'docs' in options:
+                    docs = json.loads(options['docs'])
+            for doc in docs:
+                self.indexer.add(doc)
         cherrypy.response.status = httplib.ACCEPTED
     @cherrypy.expose
     @cherrypy.tools.allow(methods=['GET', 'HEAD', 'DELETE'])
@@ -447,11 +574,16 @@ class WebIndexer(WebSearcher):
         """
         if cherrypy.request.method != 'DELETE':
             return WebSearcher.search(self, q, **options)
-        self.indexer.delete(self.parse(q, **options))
+        if q is None:
+            self.indexer.deleteAll()
+        else:
+            self.indexer.delete(self.parse(self.searcher, q, **options))
         cherrypy.response.status = httplib.ACCEPTED
     @cherrypy.expose
+    @cherrypy.tools.json(process_body=dict)
     @cherrypy.tools.allow(methods=['GET', 'HEAD', 'PUT'])
-    def fields(self, name='', **params):
+    @cherrypy.tools.validate(on=False)
+    def fields(self, name='', **settings):
         """Return or store a field's parameters.
         
         **GET** /fields
@@ -462,27 +594,26 @@ class WebIndexer(WebSearcher):
         **GET, PUT** /fields/*chars*
             Set and return parameters for given field name.
             
-            store=\ *chars*
-            
-            index=\ *chars*
-            
-            termvector=\ *chars*
+            {"store"|"index"|"termvector": *string*\|true|false,... }
             
             :return: {"store": *string*, "index": *string*, "termvector": *string*}
         """
-        if cherrypy.request.method == 'PUT':
-            self.indexer.set(name, **params)
         if not name:
+            allow()
             return sorted(self.indexer.fields)
+        if cherrypy.request.method == 'PUT':
+            if name not in self.indexer.fields:
+                cherrypy.response.status = httplib.CREATED
+            self.indexer.set(name, **settings)
         with HTTPError(httplib.NOT_FOUND, KeyError):
             field = self.indexer.fields[name]
         return dict((name, str(getattr(field, name))) for name in ['store', 'index', 'termvector'])
 
-def start(root=None, path='', config=None, pidfile='', daemonize=False, autoreload=0, autorefresh=0, callback=None):
+def start(root=None, path='', config=None, pidfile='', daemonize=False, autoreload=0, autoupdate=0, callback=None, autorefresh=0):
     """Attach root, subscribe to plugins, and start server.
     
     :param root,path,config: see cherrypy.quickstart
-    :param pidfile,daemonize,autoreload,autorefresh: see command-line options
+    :param pidfile,daemonize,autoreload,autoupdate: see command-line options
     :param callback: optional callback function scheduled after daemonizing
     """
     cherrypy.engine.subscribe('start_thread', attach_thread)
@@ -497,7 +628,10 @@ def start(root=None, path='', config=None, pidfile='', daemonize=False, autorelo
     if autoreload:
         Autoreloader(cherrypy.engine, autoreload).subscribe()
     if autorefresh:
-        AttachedMonitor(cherrypy.engine, getattr(root, 'commit', root.refresh), autorefresh).subscribe()
+        warnings.warn('Autorefresh has been renamed autoupdate.', DeprecationWarning)
+        autoupdate = autorefresh
+    if autoupdate:
+        AttachedMonitor(cherrypy.engine, root.update, autoupdate).subscribe()
     if callback:
         priority = (cherrypy.process.plugins.Daemonizer.start.priority + cherrypy.server.start.priority) // 2
         cherrypy.engine.subscribe('start', callback, priority)
@@ -509,16 +643,14 @@ parser.add_option('-c', '--config', help='optional configuration file or json ob
 parser.add_option('-p', '--pidfile', metavar='FILE', help='store the process id in the given file')
 parser.add_option('-d', '--daemonize', action='store_true', help='run the server as a daemon')
 parser.add_option('--autoreload', type=int, metavar='SECONDS', help='automatically reload modules; replacement for engine.autoreload')
-parser.add_option('--autorefresh', type=int, metavar='SECONDS', help='automatically refresh index')
+parser.add_option('--autoupdate', type=int, metavar='SECONDS', help='automatically update index version')
+parser.add_option('--autorefresh', type=int, metavar='SECONDS', help='deprecated; use autoupdate')
 
 if __name__ == '__main__':
     options, args = parser.parse_args()
     read_only = options.__dict__.pop('read_only')
     if options.config and not os.path.exists(options.config):
         options.config = {'global': json.loads(options.config)}
-    directories = list(map(os.path.abspath, args))
-    root = object.__new__(WebSearcher if (read_only or len(args) > 1) else WebIndexer)
-    def init():
-        lucene.initVM(vmargs='-Xrs')
-        root.__init__(*directories)
-    start(root, callback=init, **options.__dict__)
+    cls = WebSearcher if (read_only or len(args) > 1) else WebIndexer
+    root = cls.new(*map(os.path.abspath, args))
+    start(root, callback=root.init, **options.__dict__)
