@@ -1,6 +1,7 @@
 from future_builtins import map
 import unittest
 import heapq
+import time
 import socket, httplib
 from lupyne import client
 import local, remote
@@ -16,7 +17,7 @@ class TestCase(remote.BaseTest):
         "Distributed reading and writing."
         resources = client.Resources(self.hosts, limit=1)
         assert resources.unicast('GET', '/')
-        assert not resources.unicast('POST', '/')
+        assert not resources.unicast('POST', '/terms')
         responses = resources.broadcast('GET', '/')
         assert len(responses) == len(resources)
         for response in responses:
@@ -27,11 +28,11 @@ class TestCase(remote.BaseTest):
         responses = resources.broadcast('PUT', '/fields/name', {'store': 'yes', 'index': 'not_analyzed'})
         assert all(response() == {'index': 'NOT_ANALYZED', 'store': 'YES', 'termvector': 'NO'} for response in responses)
         doc = {'name': 'sample', 'text': 'hello world'}
-        responses = resources.broadcast('POST', '/docs', {'docs': [doc]})
+        responses = resources.broadcast('POST', '/docs', [doc])
         assert all(response() is None for response in responses)
-        response = resources.unicast('POST', '/docs', {'docs': [doc]})
+        response = resources.unicast('POST', '/docs', [doc])
         assert response() is None
-        responses = resources.broadcast('POST', '/commit')
+        responses = resources.broadcast('POST', '/update')
         assert all(response() >= 1 for response in responses)
         responses = resources.broadcast('GET', '/search?q=text:hello')
         docs = []
@@ -41,9 +42,18 @@ class TestCase(remote.BaseTest):
             docs += result['docs']
         assert len(docs) == len(resources) + 1
         assert len(set(doc['__id__'] for doc in docs)) == 2
-        self.stop(self.servers.pop())
+        self.stop(self.servers.pop(0))
         self.assertRaises((socket.error, httplib.BadStatusLine), resources.broadcast, 'GET', '/')
         assert resources.unicast('GET', '/')()
+        del resources[self.hosts[0]]
+        assert all(resources.broadcast('GET', '/'))
+        assert list(map(len, resources.values())) == [1, 1]
+        time.sleep(self.config['server.socket_timeout'] + 1)
+        assert resources.unicast('GET', '/')
+        counts = list(map(len, resources.values()))
+        assert set(counts) == set([0, 1])
+        assert resources.broadcast('GET', '/')
+        assert list(map(len, resources.values())) == counts[::-1]
         resources.clear()
         self.assertRaises(ValueError, resources.unicast, 'GET', '/')
     
@@ -53,8 +63,8 @@ class TestCase(remote.BaseTest):
         shards = client.Shards(zip(self.hosts * 2, heapq.merge(keys, keys)), limit=1)
         shards.resources.broadcast('PUT', '/fields/zone', {'store': 'yes'})
         for zone in range(len(self.ports)):
-            shards.broadcast(zone, 'POST', '/docs', {'docs': [{'zone': str(zone)}]})
-        shards.resources.broadcast('POST', '/commit')
+            shards.broadcast(zone, 'POST', '/docs', [{'zone': str(zone)}])
+        shards.resources.broadcast('POST', '/update')
         result = shards.unicast(0, 'GET', '/search?q=zone:0')()
         assert result['count'] == len(result['docs']) == 1
         assert all(response() == result for response in shards.broadcast(0, 'GET', '/search?q=zone:0'))
@@ -70,8 +80,8 @@ class TestCase(remote.BaseTest):
             assert len(docs) == 2
             zones.update(doc['zone'] for doc in docs)
         assert zones == set('012')
-        self.stop(self.servers.pop())
-        self.assertRaises((socket.error, httplib.BadStatusLine), shards.broadcast, 2, 'GET', '/')
+        self.stop(self.servers.pop(0))
+        self.assertRaises((socket.error, httplib.BadStatusLine), shards.broadcast, 0, 'GET', '/')
         responses = shards.multicast([0, 1, 2], 'GET', '/')
         assert len(responses) == 2 and all(response() for response in responses)
 
