@@ -7,6 +7,7 @@ import collections
 import warnings
 import datetime
 import math
+from contextlib import contextmanager
 import lucene
 from lupyne import engine
 import fixture
@@ -25,6 +26,20 @@ else:
             self.payload = self.type.encode('utf8')
             return result
 
+@contextmanager
+def assertWarns(*categories):
+    with warnings.catch_warnings(record=True) as messages:
+        yield
+    assert len(messages) == len(categories), messages
+    for message, category in zip(messages, categories):
+         assert issubclass(message.category, category), message
+
+class Filter(lucene.PythonFilter):
+    "Broken filter to test errors are raised."
+    def getBitSet(self, indexReader):
+        assert False
+    getDocIdSet = getBitSet
+
 class BaseTest(unittest.TestCase):
     def setUp(self):
         self.tempdir = tempfile.mkdtemp(dir=os.path.dirname(__file__))
@@ -38,13 +53,15 @@ class TestCase(BaseTest):
         self.assertRaises(TypeError, engine.IndexSearcher)
         analyzer = lucene.StandardAnalyzer(lucene.Version.LUCENE_CURRENT)
         stemmer = engine.Analyzer(analyzer, lucene.PorterStemFilter, typeAsPayload)
-        token = stemmer.tokens('hello').next()
-        assert token.positionIncrement == 1
-        if not isinstance(token, lucene.Token):
-            assert engine.TokenFilter(lucene.EmptyTokenStream()).payload is None
-            assert token.term == 'hello'
-            assert token.type == token.payload == '<ALPHANUM>'
-            assert token.offset == (0, 5)
+        for token in stemmer.tokens('hello'):
+            assert token.positionIncrement == 1
+            if not isinstance(token, lucene.Token):
+                assert engine.TokenFilter(lucene.EmptyTokenStream()).payload is None
+                assert token.term == 'hello'
+                assert token.type == token.payload == '<ALPHANUM>'
+                assert token.offset == (0, 5)
+                token.term = token.type = ''
+                token.offset, token.positionIncrement = (0, 0), 0
         assert str(stemmer.parse('hellos', field=['body', 'title'])) == 'body:hello title:hello'
         assert str(stemmer.parse('hellos', field={'body': 1.0, 'title': 2.0})) == 'body:hello title:hello^2.0'
         indexer = engine.Indexer(analyzer=stemmer)
@@ -70,6 +87,7 @@ class TestCase(BaseTest):
         doc = indexer[0]
         assert len(doc) == 3
         assert 'name' in doc and 'missing' not in doc
+        assert sorted(doc) == ['name', 'tag', 'tag']
         assert sorted(doc.items()) == [('name', 'sample'), ('tag', 'python'), ('tag', 'search')]
         assert doc.dict('tag') == {'name': 'sample', 'tag': ['python', 'search']}
         assert doc.dict(name=None, missing=True) == {'name': 'sample', 'missing': True}
@@ -101,8 +119,17 @@ class TestCase(BaseTest):
         assert not indexer.search('hello') and indexer.search('hello', field='text')
         assert indexer.search('text:hello hi') and not indexer.search('text:hello hi', op='and')
         assert indexer.search('text:*hello', allowLeadingWildcard=True)
+        query = engine.Query.multiphrase('text', ('hello', 'hi'), None, 'world')
+        assert str(query) == 'text:"(hello hi) world"' and list(query.positions) == [0, 2]
+        query = engine.Query.wildcard('text', '*')
+        assert str(query) == 'text:*' and isinstance(query, lucene.WildcardQuery)
+        assert str(lucene.MatchAllDocsQuery() | query) == '*:* text:*'
+        assert str(lucene.MatchAllDocsQuery() - query) == '*:* -text:*'
+        query = +query
+        query &= engine.Query.fuzzy('text', 'hello')
+        query |= engine.Query.fuzzy('text', 'hello', 0.1)
+        assert str(query) == '+text:* +text:hello~0.5 text:hello~0.1'
         query = engine.Query.span('text', 'world')
-        self.assertRaises(AssertionError, query.near, option=None)
         assert str(query.mask('name')) == 'mask(text:world) as name'
         query = engine.Query.disjunct(0.1, query, name='sample')
         assert str(query) == '(text:world | name:sample)~0.1'
@@ -124,6 +151,15 @@ class TestCase(BaseTest):
         indexer.optimize()
         assert len(indexer) == sum(indexer.segments.values())
         assert 0 not in indexer
+        indexer.add(tag='test', name='old')
+        indexer.update('tag', boost=2.0, tag='test')
+        indexer.commit()
+        assert [indexer[id].dict() for id in indexer] == [{'tag': 'test'}]
+        indexer.update('tag', 'test', {'name': 'new'})
+        indexer.commit()
+        assert [indexer[id].dict() for id in indexer] == [{'name': 'new'}]
+        indexer.deleteAll()
+        indexer.commit()
         temp = engine.Indexer(self.tempdir)
         temp.add()
         temp.commit()
@@ -134,6 +170,20 @@ class TestCase(BaseTest):
         indexer.add(text=lucene.WhitespaceTokenizer(lucene.StringReader('?')))
         indexer.commit()
         assert list(indexer.terms('text')) == ['?']
+        reader = engine.indexers.IndexReader(indexer.indexReader)
+        assert reader[0].dict() == {} and reader.count('text', '?') == 1
+        assert len(reader.comparator('text')) == 4
+        indexer.delete('text', '?')
+        indexer.commit(expunge=True)
+        assert not indexer.hasDeletions()
+        indexer.commit(optimize=2)
+        len(indexer.segments) <= 2
+        indexer.commit(optimize=True)
+        assert indexer.optimized
+        del reader.indexReader
+        self.assertRaises(AttributeError, getattr, reader, 'maxDoc')
+        del indexer.indexSearcher
+        self.assertRaises(AttributeError, getattr, indexer, 'search')
     
     def testBasic(self):
         "Text fields and simple searches."
@@ -145,18 +195,28 @@ class TestCase(BaseTest):
             indexer.add(doc, boost=('article' in doc) + 1.0)
         indexer.commit()
         searcher = engine.IndexSearcher.load(self.tempdir)
+        engine.IndexSearcher.load(searcher.directory) # ensure directory isn't closed
         assert len(indexer) == len(searcher) and lucene.RAMDirectory.instance_(searcher.directory)
         assert indexer.filters == indexer.spellcheckers == {}
-        assert indexer.facets([], 'amendment') and indexer.suggest('amendment', '')
+        with assertWarns(DeprecationWarning):
+            assert indexer.facets([0], 'amendment')
+            assert not indexer.facets(lucene.OpenBitSet())
+            assert not indexer.search(filter=[])
+            f = engine.Filter([])
+            assert not engine.Filter.overlap(f, f)
+        assert indexer.suggest('amendment', '')
         assert list(indexer.filters) == list(indexer.spellcheckers) == ['amendment']
         indexer.delete('amendment', doc['amendment'])
         indexer.add(doc)
         indexer.commit(filters=True, spellcheckers=True)
         assert list(indexer.filters) == list(indexer.spellcheckers) == ['amendment']
+        doc['amendment'] = engine.Analyzer(lucene.WhitespaceTokenizer).tokens(doc['amendment'])
+        scores = list(searcher.match(doc, 'text:congress', 'text:law', 'amendment:27'))
+        assert 0.0 == scores[0] < scores[1] < scores[2] < 1.0
         searcher = engine.MultiSearcher([indexer.directory, self.tempdir])
         assert searcher.count() == len(searcher) == 2 * len(indexer)
         searcher = searcher.reopen()
-        assert searcher.facets(searcher, 'amendment')['amendment'] == dict.fromkeys(map(str, range(1, 28)), 2)
+        assert searcher.facets(lucene.MatchAllDocsQuery(), 'amendment')['amendment'] == dict.fromkeys(map(str, range(1, 28)), 2)
         reader = searcher.indexReader
         del searcher
         self.assertRaises(lucene.JavaError, reader.isCurrent)
@@ -213,8 +273,10 @@ class TestCase(BaseTest):
         hits = indexer.search('text:right')
         for name in ('amendment', 'article'):
             indexer.filters[name] = engine.Query.prefix(name, '').filter()
-        assert indexer.facets(hits.ids, 'amendment', 'article') == {'amendment': 12, 'article': 1}
-        hits = indexer.search('text:people', filter=hits.ids)
+        query = engine.Query.term('text', 'right')
+        assert indexer.facets(str(query), 'amendment', 'article') == {'amendment': 12, 'article': 1}
+        self.assertRaises(lucene.InvalidArgsError, indexer.overlap, query.filter(), query.filter(cache=False))
+        hits = indexer.search('text:people', filter=query.filter())
         assert len(hits) == 4
         hit, = indexer.search('date:192*')
         assert hit['amendment'] == '19'
@@ -230,10 +292,6 @@ class TestCase(BaseTest):
         assert set(hit.get('amendment') for hit in hits) > set(amendments)
         hit, = indexer.search(query & engine.Query.term('text', 'vote'))
         assert hit['amendment'] == '19'
-        f = engine.Filter(indexer.docs('text', 'vote'))
-        for n in range(2):  # filters should be reusable
-            hit, = indexer.search(query, filter=f)
-            assert hit['amendment'] == '19'
         hit, = indexer.search(query - engine.Query.all(text='vote'))
         assert hit['amendment'] == '18'
         hit, = indexer.search(engine.Query.all(text=['persons', 'papers']))
@@ -291,6 +349,7 @@ class TestCase(BaseTest):
         assert lucene.TermQuery.instance_(query) and str(query) == 'text:writs'
         query = indexer.parse('"hello world"', field='text', spellcheck=True)
         assert lucene.PhraseQuery.instance_(query) and str(query) == 'text:"held would"'
+        assert str(indexer.parse('vwxyz', field='text', spellcheck=True)) == 'text:vwxyz'
         del indexer
         assert engine.Indexer(self.tempdir)
     
@@ -299,8 +358,10 @@ class TestCase(BaseTest):
         indexer = engine.Indexer(self.tempdir)
         for name, params in fixture.zipcodes.fields.items():
             indexer.set(name, **params)
-        indexer.set('longitude', engine.PrefixField, store=True)
-        indexer.fields['location'] = engine.NestedField('state.county.city', sep='.')
+        with assertWarns(DeprecationWarning, DeprecationWarning):
+            indexer.set('longitude', engine.PrefixField, store=True)
+            assert engine.NestedField('state:county:city').sep == ':'
+        indexer.fields['location'] = engine.NestedField('state.county.city')
         for doc in fixture.zipcodes.docs():
             if doc['state'] in ('CA', 'AK', 'WY', 'PR'):
                 lat, lng = ('{0:08.3f}'.format(doc.pop(l)) for l in ['latitude', 'longitude'])
@@ -340,19 +401,19 @@ class TestCase(BaseTest):
         assert count == indexer.count(field.range(lng[:3], lng[:3]+'~'))
         assert count > indexer.count(engine.Query.term('state', 'CA'), filter=engine.Query.term(longitude, lng).filter())
         query = engine.Query.prefix('zipcode', '90')
-        (field, facets), = indexer.facets(indexer.search(query).ids, 'state.county').items()
-        assert [facets] == indexer.facets(query.filter(), 'state.county').values()
+        (field, facets), = indexer.facets(query, 'state.county').items()
         assert field == 'state.county'
         la, orange = sorted(filter(facets.get, facets))
         assert la == 'CA.Los Angeles' and facets[la] > 100
         assert orange == 'CA.Orange' and facets[orange] > 10
-        (field, facets), = indexer.facets(query.filter(), ('state.county', 'CA.*')).items()
+        (field, facets), = indexer.facets(query, ('state.county', 'CA.*')).items()
         assert all(value.startswith('CA.') for value in facets) and set(facets) < set(indexer.filters['state.county'])
         for count in (None, len(indexer)):
             hits = indexer.search(query, count=count, timeout=0.01)
             assert 0 <= len(hits) <= indexer.count(query) and hits.count in (None, len(hits)) and hits.maxscore in (None, 1.0)
             hits = indexer.search(query, count=count, timeout=-1)
             assert len(hits) == 0 and hits.count is hits.maxscore is None
+        self.assertRaises(lucene.JavaError, indexer.search, filter=Filter())
         directory = lucene.RAMDirectory()
         query = engine.Query.term('state', 'CA')
         size = indexer.copy(directory, query)
@@ -366,57 +427,51 @@ class TestCase(BaseTest):
         directory.close()
     
     def testSpatial(self):
-        "Spatial tile test."
-        for PolygonField in (engine.PolygonField, engine.numeric.PolygonField):
-            PointField, = PolygonField.__bases__
-            indexer = engine.Indexer(self.tempdir, 'w')
-            for name, params in fixture.zipcodes.fields.items():
-                indexer.set(name, **params)
-            with warnings.catch_warnings(record=True) as deprecations:
-                field = indexer.fields['tile'] = PointField('tile', precision=15, step=2, store=True)
-            assert deprecations or isinstance(field, engine.numeric.PointField)
-            points = []
-            for doc in fixture.zipcodes.docs():
-                if doc['state'] == 'CA':
-                    lat, lng = doc.pop('latitude'), doc.pop('longitude')
-                    indexer.add(doc, tile=[(lng, lat)], latitude=str(lat), longitude=str(lng))
-                    if doc['city'] == 'Los Angeles':
-                        points.append((lng, lat))
-            assert len(list(PolygonField('', precision=15).items(points))) > len(points)
-            indexer.commit()
-            city, zipcode, tile = 'Beverly Hills', '90210', '023012311120332'
-            hit, = indexer.search('zipcode:' + zipcode)
-            assert (hit['tile'] == tile or int(hit['tile']) == int(tile, 4)) and hit['city'] == city
-            hit, = indexer.search(field.prefix(tile))
-            assert hit['zipcode'] == zipcode and hit['city'] == city
-            x, y = (float(hit[l]) for l in ['longitude', 'latitude'])
-            assert field.coords(tile[:4]) == (2, 9)
-            bottom, left, top, right = field.decode(tile)
-            assert left < x < right and bottom < y < top
-            hits = indexer.search(field.near(x, y))
-            cities = set(hit['city'] for hit in hits)
-            assert set([city]) == cities
-            hits = indexer.search(field.near(x, y, precision=10))
-            cities = set(hit['city'] for hit in hits)
-            assert city in cities and len(cities) > 10
-            query = field.within(x, y, 10**4)
-            assert issubclass(PointField, engine.PointField) or len(query) < 3
-            cities = set(hit['city'] for hit in indexer.search(query))
-            assert city in cities and 100 > len(cities) > 50
-            hits = indexer.search(field.within(x, y, 10**5))
-            cities = set(hit['city'] for hit in hits)
-            assert city in cities and len(cities) > 100
-            assert len(field.within(x, y, 10**8)) == 1
-            del indexer
+        "Spatial tiles."
+        indexer = engine.Indexer(self.tempdir, 'w')
+        for name, params in fixture.zipcodes.fields.items():
+            indexer.set(name, **params)
+        field = indexer.fields['tile'] = engine.PointField('tile', precision=15, step=2, store=True)
+        points = []
+        for doc in fixture.zipcodes.docs():
+            if doc['state'] == 'CA':
+                lat, lng = doc.pop('latitude'), doc.pop('longitude')
+                indexer.add(doc, tile=[(lng, lat)], latitude=str(lat), longitude=str(lng))
+                if doc['city'] == 'Los Angeles':
+                    points.append((lng, lat))
+        assert len(list(engine.PolygonField('', precision=15).items(points))) > len(points)
+        indexer.commit()
+        city, zipcode, tile = 'Beverly Hills', '90210', '023012311120332'
+        hit, = indexer.search('zipcode:' + zipcode)
+        assert (hit['tile'] == tile or int(hit['tile']) == int(tile, 4)) and hit['city'] == city
+        hit, = indexer.search(field.prefix(tile))
+        assert hit['zipcode'] == zipcode and hit['city'] == city
+        x, y = (float(hit[l]) for l in ['longitude', 'latitude'])
+        assert field.coords(tile[:4]) == (2, 9)
+        bottom, left, top, right = field.decode(tile)
+        assert left < x < right and bottom < y < top
+        hits = indexer.search(field.near(x, y))
+        cities = set(hit['city'] for hit in hits)
+        assert set([city]) == cities
+        hits = indexer.search(field.near(x, y, precision=10))
+        cities = set(hit['city'] for hit in hits)
+        assert city in cities and len(cities) > 10
+        query = field.within(x, y, 10**4)
+        assert len(query) < 3
+        cities = set(hit['city'] for hit in indexer.search(query))
+        assert city in cities and 100 > len(cities) > 50
+        hits = indexer.search(field.within(x, y, 10**5))
+        cities = set(hit['city'] for hit in hits)
+        assert city in cities and len(cities) > 100
+        assert len(field.within(x, y, 10**8)) == 1
+        del indexer
     
     def testFields(self):
-        "Custom field tests."
+        "Custom fields."
         indexer = engine.Indexer(self.tempdir)
         indexer.set('amendment', engine.FormatField, format='{0:02d}', store=True)
-        with warnings.catch_warnings(record=True) as deprecations:
-            indexer.set('date', engine.DateTimeField, store=True)
-        assert deprecations
         indexer.set('size', engine.FormatField, format='{0:04d}', store=True)
+        field = indexer.fields['date'] = engine.NestedField('Y-m-d', sep='-', store=True)
         for doc in fixture.constitution.docs():
             if 'amendment' in doc:
                 indexer.add(amendment=int(doc['amendment']), date=doc['date'], size=len(doc['text']))
@@ -425,26 +480,16 @@ class TestCase(BaseTest):
         assert indexer.count(query) == 9
         query = engine.Query.prefix('amendment', '0')
         assert indexer.count(query) == 9
-        field = indexer.fields['date']
         query = field.prefix('1791-12-15')
         assert indexer.count(query) == 10
         query = field.range('', '1921-12', lower=False, upper=True)
-        assert str(query) == 'date:Y:{ TO 1921} date:Ym:[1921 TO 1921-12]'
+        assert str(query) == 'Y-m:{ TO 1921-12]', query
         assert indexer.count(query) == 19
-        query = field.range(datetime.date(1919, 1, 1), datetime.date(1921, 12, 31))
-        fields = [lucene.TermRangeQuery.cast_(clause.query).field for clause in query]
-        assert fields == ['date:Ymd', 'date:Ym', 'date:Y', 'date:Ym', 'date:Ymd']
+        query = field.range('1919-01-01', '1921-12-31')
+        assert str(query) == 'Y-m-d:[1919-01-01 TO 1921-12-31}'
         hits = indexer.search(query)
         assert [hit['amendment'] for hit in hits] == ['18', '19']
-        assert [hit['date'].split('-')[0] for hit in hits] == ['1919', '1920']
-        query = field.within(seconds=100)
-        assert indexer.count(query) == 0
-        query = field.duration([2009], days=-100*365)
-        assert 0 < len(query) <= 5
-        assert indexer.count(query) == 12
-        assert len(field.within(-100)) <= 3
-        assert len(field.within(-100.0)) > 3
-        assert len(field.within(-100, seconds=1, utc=True)) > 3
+        assert [hit['Y-m-d'].split('-')[0] for hit in hits] == ['1919', '1920']
         field = indexer.fields['size']
         sizes = dict((id, int(indexer[id]['size'])) for id in indexer)
         ids = sorted((id for id in sizes if sizes[id] >= 1000), key=sizes.get)
@@ -455,7 +500,7 @@ class TestCase(BaseTest):
         assert hits.ids == ids[:len(hits)]
         query = engine.Query.range('size', None, '1000')
         assert indexer.count(query) == len(sizes) - len(ids)
-        indexer.sorters['year'] = engine.SortField('date:Ymd', type=int, parser=lambda date: int(date.split('-')[0]))
+        indexer.sorters['year'] = engine.SortField('Y-m-d', type=int, parser=lambda date: int(date.split('-')[0]))
         assert indexer.comparator('year')[:10] == [1791] * 10
         hits = indexer.search(count=3, sort='year')
         assert [int(hit['amendment']) for hit in hits] == [1, 2, 3]
@@ -469,12 +514,13 @@ class TestCase(BaseTest):
         assert indexer.comparator('year')[-1] == 0
         assert cache == len(lucene.FieldCache.DEFAULT.cacheEntries)
     
-    def testNumericFields(self):
-        "Numeric variant fields."
+    def testNumeric(self):
+        "Numeric fields."
         indexer = engine.Indexer(self.tempdir)
-        indexer.set('amendment', engine.numeric.NumericField, store=True)
-        indexer.set('date', engine.numeric.DateTimeField, store=True)
-        indexer.set('size', engine.numeric.NumericField, store=True, step=5)
+        with assertWarns(DeprecationWarning):
+            indexer.set('amendment', engine.numeric.NumericField, store=True)
+            indexer.set('date', engine.numeric.DateTimeField, store=True)
+            indexer.set('size', engine.numeric.NumericField, store=True, step=5)
         for doc in fixture.constitution.docs():
             if 'amendment' in doc:
                 indexer.add(amendment=int(doc['amendment']), date=[tuple(map(int, doc['date'].split('-')))], size=len(doc['text']))
@@ -492,8 +538,7 @@ class TestCase(BaseTest):
         hits = indexer.search(query)
         assert [hit['amendment'] for hit in hits] == ['18', '19']
         assert [datetime.datetime.utcfromtimestamp(float(hit['date'])).year for hit in hits] == [1919, 1920]
-        query = field.within(seconds=100)
-        assert indexer.count(query) == 0
+        assert indexer.count(field.within(seconds=100)) == indexer.count(field.within(weeks=1)) == 0
         query = field.duration([2009], days=-100*365)
         assert indexer.count(query) == 12
         field = indexer.fields['size']
@@ -543,5 +588,5 @@ class TestCase(BaseTest):
         assert fragment.count('<B>') == fragment.count('</B>') == 3
 
 if __name__ == '__main__':
-    lucene.initVM(lucene.CLASSPATH)
+    lucene.initVM()
     unittest.main()
