@@ -2,16 +2,18 @@ from future_builtins import map
 import unittest
 import warnings
 import os, sys
-import subprocess
+import signal, subprocess
 import operator
 import httplib
 import math
 import json
+import time, calendar
 import socket, errno
 from contextlib import contextmanager
+from email.utils import parsedate
 import lucene
 import cherrypy
-from lupyne import client, server
+from lupyne import client, engine, server
 import fixture, local
 
 @contextmanager
@@ -26,28 +28,25 @@ def assertRaises(exception, code):
 
 class BaseTest(local.BaseTest):
     ports = 8080, 8081
+    config = {'server.socket_timeout': 2, 'server.shutdown_timeout': 1}
+    def run(self, result):
+        self.config['log.screen'] = result.showAll
+        local.BaseTest.run(self, result)
     def setUp(self):
         local.BaseTest.setUp(self)
-        pidfile = os.path.join(self.tempdir, 'pid')
-        self.servers = (
-            self.start(self.ports[0], self.tempdir, '--autoreload=1'),
-            self.start(self.ports[1], self.tempdir, self.tempdir, '-p', pidfile), # concurrent searchers
-        )
-        assert int(open(pidfile).read()) == self.servers[-1].pid
-    def run(self, result):
-        self.verbose = result.showAll
-        local.BaseTest.run(self, result)
+        self.servers = []
     def tearDown(self):
         for server in self.servers:
             self.stop(server)
         local.BaseTest.tearDown(self)
-    def start(self, port, *args):
+    def start(self, port, *args, **config):
         "Start server in separate process on given port."
-        config = json.dumps({'server.socket_port': port, 'log.screen': self.verbose})
+        config.update(self.config)
+        config['server.socket_port'] = port
         cherrypy.process.servers.wait_for_free_port('localhost', port)
-        server = subprocess.Popen((sys.executable, '-m', 'lupyne.server', '-c', config) + args)
+        server = subprocess.Popen((sys.executable, '-m', 'lupyne.server', '-c', json.dumps(config)) + args)
         cherrypy.process.servers.wait_for_occupied_port('localhost', port)
-        assert server.poll() is None
+        assert not server.poll()
         return server
     def stop(self, server):
         "Terminate server."
@@ -58,22 +57,40 @@ class TestCase(BaseTest):
     
     def testInterface(self):
         "Remote reading and writing."
+        self.servers += (
+            self.start(self.ports[0], self.tempdir, '--autoreload=1', **{'tools.validate.expires': 0, 'tools.validate.max_age': 0}),
+            self.start(self.ports[1], self.tempdir, self.tempdir, '--autoupdate=2'), # concurrent searchers
+        )
         resource = client.Resource('localhost', self.ports[0])
         assert resource.get('/favicon.ico')
-        resource.request('GET', '/')
-        response = resource.getresponse()
+        response = resource.call('GET', '/')
         assert response.status == httplib.OK and response.reason == 'OK' and response.time > 0
-        assert response.getheader('content-encoding') == 'gzip' and response.getheader('content-type').startswith('text/x-json')
+        assert response.getheader('content-encoding') == 'gzip' and response.getheader('content-type').startswith('application/json')
+        version, modified = response.getheader('etag'), response.getheader('last-modified')
+        assert version.strip('W/"').isdigit()
+        assert int(response.getheader('age')) >= 0 and response.getheader('cache-control') == 'max-age=0'
+        dates = list(map(parsedate, [modified, response.getheader('expires'), response.getheader('date')]))
+        assert all(dates) and sorted(dates) == dates
         (directory, count), = response().items()
         assert count == 0 and 'FSDirectory@' in directory
         assert resource.call('HEAD', '/').status == httplib.OK
         with assertRaises(httplib.HTTPException, httplib.METHOD_NOT_ALLOWED):
-            resource.put('/')
+            resource.post('/terms')
         with assertRaises(httplib.HTTPException, httplib.METHOD_NOT_ALLOWED):
-            resource.post('/fields')
+            resource.get('/update')
+        with assertRaises(httplib.HTTPException, httplib.METHOD_NOT_ALLOWED):
+            resource.put('/fields')
+        with assertRaises(httplib.HTTPException, httplib.METHOD_NOT_ALLOWED):
+            resource.post('/docs/x/y')
+        with assertRaises(httplib.HTTPException, httplib.METHOD_NOT_ALLOWED):
+            resource.put('/docs/0')
+        with assertRaises(httplib.HTTPException, httplib.METHOD_NOT_ALLOWED):
+            resource.delete('/docs')
         assert resource.get('/docs') == []
         with assertRaises(httplib.HTTPException, httplib.NOT_FOUND):
             resource.get('/docs/0')
+        with assertRaises(httplib.HTTPException, httplib.NOT_FOUND):
+            resource.get('/docs/x/y')
         try:
             assert resource.get('/docs/~')
         except httplib.HTTPException as exc:
@@ -91,26 +108,48 @@ class TestCase(BaseTest):
         assert resource.get('/terms/x/y/docs') == []
         assert resource.get('/terms/x/y/docs/counts') == []
         assert resource.get('/terms/x/y/docs/positions') == []
-        assert resource.put('/fields/text') == {'index': 'ANALYZED', 'store': 'NO', 'termvector': 'NO'}
-        assert resource.put('/fields/name', store='yes', index='not_analyzed')
+        with assertRaises(httplib.HTTPException, httplib.NOT_FOUND):
+            resource.get('/terms/x/y/missing')
+        with assertRaises(httplib.HTTPException, httplib.NOT_FOUND):
+            resource.get('/terms/x/y/docs/missing')
+        defaults = {'index': 'ANALYZED', 'store': 'NO', 'termvector': 'NO'}
+        response = resource.call('PUT', '/fields/text')
+        assert response.getheader('etag') is None
+        assert response.status == httplib.CREATED and response() == defaults
+        response = resource.call('PUT', '/fields/text', {})
+        assert response.status == httplib.OK and response() == defaults
+        assert resource.put('/fields/name', {'store': True, 'index': 'not_analyzed'})
         assert sorted(resource.get('/fields')) == ['name', 'text']
         assert resource.get('/fields/text')['index'] == 'ANALYZED'
-        assert not resource.post('/docs', docs=[{'name': 'sample', 'text': 'hello world'}])
-        with assertRaises(httplib.HTTPException, httplib.BAD_REQUEST):
-            resource.post('/docs', docs='')
+        assert not resource.post('/docs', [{'name': 'sample', 'text': 'hello world'}])
+        assert not resource.post('/docs')
         (directory, count), = resource.get('/').items()
         assert count == 1
         assert resource.get('/docs') == []
         result = resource.get('/search?q=text:hello')
         assert math.isnan(result.pop('maxscore'))
         assert result == {'query': 'text:hello', 'count': 0, 'docs': []}
-        assert resource.post('/commit')
+        resource.headers['if-none-match'] = version
+        response = resource.call('GET', '/', redirect=True)
+        assert response.status == httplib.NOT_MODIFIED and response.getheader('etag') == version
+        del resource.headers['if-none-match']
+        resource.headers['if-modified-since'] = modified
+        assert resource.call('GET', '/').status == httplib.NOT_MODIFIED
+        time.sleep(max(0, calendar.timegm(parsedate(modified)) + 1 - time.time()))
+        assert resource.post('/update')
+        response = resource.call('GET', '/')
+        assert response and response.getheader('etag') != version and parsedate(response.getheader('last-modified')) > parsedate(modified)
+        resource.headers['if-match'] = version
+        assert resource.call('GET', '/docs/0').status == httplib.PRECONDITION_FAILED
+        del resource.headers['if-match']
         assert resource.get('/docs') == [0]
-        assert resource.get('/docs/0') == {'name': 'sample'}
+        assert resource.get('/docs/0') == resource.get('/docs/name/sample') == {'name': 'sample'}
         assert resource.get('/docs/0', fields='missing') == {'missing': None}
-        with warnings.catch_warnings(record=True) as deprecations:
+        with local.assertWarns(UserWarning, DeprecationWarning, UserWarning):
             assert resource.get('/docs/0', multifields='name') == {'name': ['sample']}
-        assert deprecations
+            assert not resource.post('/docs', docs=[])
+            httplib.HTTPConnection.request(resource, 'POST', '/docs', 'docs=[]', {'content-length': '7', 'content-type': 'application/x-www-form-urlencoded'})
+            assert not resource.getresponse()()
         assert resource.get('/docs/0', fields='', **{'fields.multi': 'missing'}) == {'missing': []}
         assert resource.get('/terms') == ['name', 'text']
         assert resource.get('/terms', option='unindexed') == []
@@ -120,6 +159,7 @@ class TestCase(BaseTest):
             resource.get('/terms/text/world~-')
         with assertRaises(httplib.HTTPException, httplib.BAD_REQUEST):
             resource.get('/terms/text/world~?count=')
+        assert resource.get('/terms/text/world?count=1')
         assert resource.get('/terms/text/world/docs') == [0]
         assert resource.get('/terms/text/world/docs/counts') == [[0, 1]]
         assert resource.get('/terms/text/world/docs/positions') == [[0, [1]]]
@@ -129,6 +169,12 @@ class TestCase(BaseTest):
             resource.get('/search', sort='-x,y')
         with assertRaises(httplib.HTTPException, httplib.BAD_REQUEST):
             resource.get('/search', count=1, sort='x:str')
+        with assertRaises(httplib.HTTPException, httplib.BAD_REQUEST):
+            resource.get('/search', count=1, group='x:str')
+        with assertRaises(httplib.HTTPException, httplib.BAD_REQUEST):
+            resource.get('/search', q='')
+        with assertRaises(httplib.HTTPException, httplib.BAD_REQUEST):
+            resource.get('/search?q.test=True')
         assert resource.get('/search', count=0) == {'count': 1, 'maxscore': 1.0, 'query': None, 'docs': []}
         assert resource.get('/search', fields='')['docs'] == [{'__id__': 0, '__score__': 1.0}]
         hit, = resource.get('/search', fields='', **{'fields.multi': 'name'})['docs']
@@ -149,44 +195,76 @@ class TestCase(BaseTest):
         assert hit['__id__'] == doc['__id__'] and hit['__score__'] > doc['__score__']
         result = resource.get('/search?q=hello+world&q.field=text^4&q.field=body')
         assert result['query'] == '(body:hello text:hello^4.0) (body:world text:world^4.0)'
+        assert resource.get('/search', q='hello', **{'q.field': 'body.title^2.0'})['query'] == 'body.title:hello^2.0'
         hit, = result['docs']
         assert hit['__id__'] == doc['__id__'] and hit['__score__'] > doc['__score__']
         result = resource.get('/search', facets='name', spellcheck=1)
         assert result['facets'] == {'name': {'sample': 1}} and result['spellcheck'] == {}
         resource = client.Resource('localhost', self.ports[-1])
-        assert resource.get('/docs') == []
-        assert resource.post('/refresh', filters=True, sorters=True) == 2
+        assert set(resource.get('/').values()) < set([0, 1])
+        assert resource.post('/update', ['filters', 'sorters']) == 2
         assert resource.get('/docs') == [0, 1]
         with assertRaises(httplib.HTTPException, httplib.NOT_FOUND):
-            resource.get('/fields')
+            resource.get('/docs/name/sample')
         with assertRaises(httplib.HTTPException, httplib.NOT_FOUND):
-            resource.post('/commit')
+            resource.get('/fields')
+        with assertRaises(httplib.HTTPException, httplib.MOVED_PERMANENTLY):
+            resource.post('/refresh', ['spellcheckers'])
+        with local.assertWarns(DeprecationWarning):
+            assert resource.call('POST', '/refresh', redirect=True)
         resource = client.Resource('localhost', self.ports[0])
         assert not resource.delete('/search', q='sample', **{'q.field': 'name', 'q.type': 'term'})
         assert resource.get('/docs') == [0]
-        assert not resource.post('/commit')
+        assert not resource.post('/update', ['expunge'])
         assert resource.get('/docs') == []
+        assert not resource.put('/docs/name/sample')
+        assert resource.post('/update')
+        assert resource.get('/docs/name/sample')
+        assert not resource.put('/docs/name/sample')
+        assert resource.post('/update')
+        assert resource.get('/docs/name/sample')
+        with assertRaises(httplib.HTTPException, httplib.CONFLICT):
+            resource.put('/docs/missing/sample')
+        with assertRaises(httplib.HTTPException, httplib.CONFLICT):
+            resource.put('/docs/missing/sample', {'name': 'mismatched'})
+        assert resource.put('/fields/name', {'index': False})
+        with assertRaises(httplib.HTTPException, httplib.CONFLICT):
+            resource.put('/docs/missing/sample')
+        assert not resource.delete('/docs/missing/sample')
+        resource.post('/update')
+        with assertRaises(httplib.HTTPException, httplib.NOT_FOUND):
+            resource.get('/docs/missing/sample')
+        assert not resource.delete('/search')
         with assertRaises(httplib.HTTPException, httplib.MOVED_PERMANENTLY):
-            resource.post('/refresh', spellcheckers=True)
-        responses = resource.multicall(('POST', '/docs', {'docs': [{}]}), ('POST', '/commit'), ('GET', '/docs'))
-        assert responses[0].status == httplib.ACCEPTED and responses[1]() == 1 and responses[2]() == [1]
+            resource.post('/commit', ['spellcheckers'])
+        responses = resource.multicall(('POST', '/docs', [{}]), ('POST', '/update'), ('GET', '/docs'))
+        assert responses[0].status == httplib.ACCEPTED and responses[1]() == len(responses[2]()) == 1
+        assert resource.post('/', [self.tempdir]).values() == [2]
         resource = client.Resource('localhost', self.ports[-1] + 1)
         with assertRaises(socket.error, errno.ECONNREFUSED):
             resource.get('/')
+        self.stop(self.servers.pop(0))
+        pidfile = os.path.join(self.tempdir, 'pid')
+        server = self.start(self.ports[0], '-dp', pidfile)
+        assert server.poll() == 0
+        os.kill(int(open(pidfile).read()), signal.SIGKILL)
     
     def testBasic(self):
         "Remote text indexing and searching."
+        with local.assertWarns(DeprecationWarning):
+            self.assertRaises(AttributeError, server.start, autorefresh=1)
+        self.servers.append(self.start(self.ports[0], self.tempdir))
         resource = client.Resource('localhost', self.ports[0])
         assert resource.get('/fields') == []
         for name, settings in fixture.constitution.fields.items():
-            assert resource.put('/fields/' + name, **settings)
+            assert resource.put('/fields/' + name, settings)
         fields = resource.get('/fields')
         assert sorted(fields) == ['amendment', 'article', 'date', 'text']
         for field in fields:
             assert sorted(resource.get('/fields/' + name)) == ['index', 'store', 'termvector']
-        resource.post('/docs/', docs=list(fixture.constitution.docs()))
+        resource.post('/docs', list(fixture.constitution.docs()))
         assert resource.get('/').values() == [35]
-        resource.post('/commit', spellcheckers=True, filters='')
+        resource.post('/update', ['optimize', 'spellcheckers'])
         assert resource.get('/docs/0', **{'fields.indexed': 'amendment:int'}) == {'amendment': 0, 'article': 'Preamble'}
         assert resource.get('/terms') == ['amendment', 'article', 'date', 'text']
         articles = resource.get('/terms/article')
@@ -294,7 +372,7 @@ class TestCase(BaseTest):
         assert result['count'] == 8 and set(doc['__score__'] for doc in result['docs']) == set([1.0])
         result = resource.get('/search', q='text:right', filter='text:people')
         assert result['count'] == 4 and 0 < result['maxscore'] < 1.0
-        result = resource.get('/search', q='text:right', group='date', **{'group.count': 2})
+        result = resource.get('/search', q='text:right', group='date', **{'group.count': 2, 'group.limit': 2})
         assert 'docs' not in result and len(result['groups']) == 9
         assert sum(map(operator.itemgetter('count'), result['groups'])) == result['count'] == 13
         assert all(min(group['count'], 2) >= len(group['docs']) for group in result['groups'])
@@ -303,11 +381,39 @@ class TestCase(BaseTest):
         assert group['value'] == '1791-12-15' and result['groups'][-1]['value'] is None
         assert sorted(group) == ['count', 'docs', 'value'] and group['count'] == 5
         assert len(group['docs']) == 2 and group['docs'][0]['amendment'] == '2'
+        assert len(result['groups'][1]['docs']) == 1 and all(group['docs'] == [] for group in result['groups'][2:])
         result = resource.get('/search', q='text:right', group='amendment:int')
         assert set(map(operator.itemgetter('count'), result['groups'])) == set([1])
         assert all(int(doc.get('amendment', 0)) == group['value'] for group in result['groups'] for doc in group['docs'])
         assert result['groups'][0]['value'] == 2 and result['groups'][-1]['value'] == 0
+    
+    def testAdvanced(self):
+        "Nested and numeric fields."
+        writer = engine.IndexWriter(self.tempdir)
+        writer.commit()
+        self.servers.append(self.start(self.ports[0], '-r', self.tempdir, **{'tools.validate.etag': False, 'tools.validate.last_modified': False}))
+        writer.set('zipcode', engine.NumericField, store=True)
+        writer.fields['location'] = engine.NestedField('county.city')
+        for doc in fixture.zipcodes.docs():
+            if doc['state'] == 'CA':
+                writer.add(zipcode=doc['zipcode'], location='{0}.{1}'.format(doc['county'], doc['city']))
+        writer.commit()
+        resource = client.Resource('localhost', self.ports[0])
+        assert resource.post('/update') == resource.get('/').popitem()[1] == len(writer)
+        terms = resource.get('/terms/zipcode:int')
+        assert len(terms) == len(writer) and terms[0] == 90001
+        terms = resource.get('/terms/zipcode:int?step=4')
+        assert len(terms) < len(writer) and terms[0] == 90000
+        result = resource.get('/search', count=0, facets='county')
+        facets = result['facets']['county']
+        assert result['count'] == sum(facets.values()) and 'Los Angeles' in facets
+        result = resource.get('/search', count=0, facets='county.city:Los Angeles.*')
+        facets = result['facets']['county.city']
+        assert result['count'] > sum(facets.values()) and all(location.startswith('Los Angeles.') for location in facets)
+        result = resource.get('/search', q='Los Angeles', group='county.city', **{'group.count': 0, 'q.field': 'county', 'q.type': 'prefix'})
+        assert all(group['value'].startswith('Los Angeles') for group in result['groups'])
+        assert sum(map(operator.itemgetter('count'), result['groups'])) == sum(facets.values()) == result['count']
 
 if __name__ == '__main__':
-    lucene.initVM(lucene.CLASSPATH)
+    lucene.initVM()
     unittest.main()
